/*
 *  Database record class for DPAP sharing
 *
 *  Copyright (C) 2008 W. Michael Petullo <mike@flyn.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <wand/wand_api.h>
#include <libexif/exif-data.h>
#include <libexif/exif-content.h>
#include <libexif/exif-entry.h>

#include "dmapd-dpap-record.h"

struct DmapdDPAPRecordPrivate {
	gint filesize;
	gint largefilesize;
	gint creationdate;
	gint rating;
	const char *location;
	const char *filename;
	const char *title;
	size_t thumbnail_size;
	unsigned char *thumbnail;
	const char *aspectratio;
	gint height;
	gint width;
	const char *format;
	const char *comments;
};

const char *dmapd_dpap_record_get_location    (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->location;
}

const char *dmapd_dpap_record_get_title       (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->title;
}

const char *dmapd_dpap_record_get_aspectratio (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->aspectratio;
}

const char *dmapd_dpap_record_get_filename (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->filename;
}

const char *dmapd_dpap_record_get_format (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->format;
}

const char *dmapd_dpap_record_get_comments (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->comments;
}

gint        dmapd_dpap_record_get_filesize (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->filesize;
}

gint        dmapd_dpap_record_get_largefilesize (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->largefilesize;
}

gint        dmapd_dpap_record_get_pixelheight (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->height;
}

gint        dmapd_dpap_record_get_pixelwidth (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->width;
}

gint        dmapd_dpap_record_get_rating (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->rating;
}

gint        dmapd_dpap_record_get_creationdate (DPAPRecord *record)
{
	return DMAPD_DPAP_RECORD (record)->priv->creationdate;
}

unsigned char *dmapd_dpap_record_get_thumbnail (DPAPRecord *record, size_t *size)
{
	g_warning ("THUMBSIZE: %d", DMAPD_DPAP_RECORD (record)->priv->thumbnail_size);
	g_warning ("THUMB: %c", *(DMAPD_DPAP_RECORD (record)->priv->thumbnail));
	*size = DMAPD_DPAP_RECORD (record)->priv->thumbnail_size;
	return DMAPD_DPAP_RECORD (record)->priv->thumbnail;
}

GInputStream *dmapd_dpap_record_read (DPAPRecord *record, GError **error)
{
        GFile *file;
        GInputStream *stream;

        file = g_file_new_for_uri (DMAPD_DPAP_RECORD (record)->priv->location);
        stream = G_INPUT_STREAM (g_file_read (file, NULL, error));

        g_object_unref (file);

        return stream;
}

GByteArray *dmapd_dpap_record_to_blob (DPAPRecord *record, size_t *size)
{
	/* FIXME: not endian safe. Don't write with one endianness and 
	 * read with another
	 */

	GByteArray *blob = g_byte_array_new ();

	*size = 0;

	/* NOTE: do not store ID in the blob. */

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->filesize),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->filesize));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->filesize);

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->largefilesize),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->largefilesize));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->largefilesize);

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->location,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->location) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->location) + 1;

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->filename,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->filename) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->filename) + 1;

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->thumbnail_size),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->thumbnail_size));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->thumbnail_size);

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->thumbnail,
			     DMAPD_DPAP_RECORD (record)->priv->thumbnail_size);
	*size += DMAPD_DPAP_RECORD (record)->priv->thumbnail_size;

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->aspectratio,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->aspectratio) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->aspectratio) + 1;

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->height),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->height));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->height);

	g_byte_array_append (blob,
			     &(DMAPD_DPAP_RECORD (record)->priv->width),
			     sizeof (DMAPD_DPAP_RECORD (record)->priv->width));
	*size += sizeof (DMAPD_DPAP_RECORD (record)->priv->width);

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->format,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->format) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->format) + 1;

	g_byte_array_append (blob,
			     DMAPD_DPAP_RECORD (record)->priv->comments,
			     strlen (DMAPD_DPAP_RECORD (record)->priv->comments) + 1);
	*size += strlen (DMAPD_DPAP_RECORD (record)->priv->comments) + 1;

	return blob;
}

static void dmapd_dpap_record_init (DmapdDPAPRecord *record)
{
	record->priv = DMAPD_DPAP_RECORD_GET_PRIVATE (record);
}

static void dmapd_dpap_record_class_init (DmapdDPAPRecordClass *klass)
{
	g_type_class_add_private (klass, sizeof (DmapdDPAPRecordPrivate));
}

static void dmapd_dpap_record_interface_init (gpointer iface, gpointer data)
{
	DPAPRecordInterface *dpap_record = iface;

	g_assert (G_TYPE_FROM_INTERFACE (dpap_record) == TYPE_DPAP_RECORD);

	dpap_record->get_location = dmapd_dpap_record_get_location;
	dpap_record->get_title = dmapd_dpap_record_get_title;
	dpap_record->get_format = dmapd_dpap_record_get_format;
	dpap_record->get_filesize = dmapd_dpap_record_get_filesize;
	dpap_record->get_rating = dmapd_dpap_record_get_rating;
	dpap_record->get_filename = dmapd_dpap_record_get_filename;
	dpap_record->get_aspectratio = dmapd_dpap_record_get_aspectratio;
	dpap_record->get_creationdate = dmapd_dpap_record_get_creationdate;
	dpap_record->get_largefilesize = dmapd_dpap_record_get_largefilesize;
	dpap_record->get_pixelheight = dmapd_dpap_record_get_pixelheight;
	dpap_record->get_pixelwidth = dmapd_dpap_record_get_pixelwidth;
	dpap_record->get_comments = dmapd_dpap_record_get_comments;
	dpap_record->get_thumbnail = dmapd_dpap_record_get_thumbnail;
	dpap_record->read = dmapd_dpap_record_read;
}

G_DEFINE_TYPE_WITH_CODE (DmapdDPAPRecord, dmapd_dpap_record, G_TYPE_OBJECT, 
			 G_IMPLEMENT_INTERFACE (TYPE_DPAP_RECORD, dmapd_dpap_record_interface_init))

DmapdDPAPRecord *dmapd_dpap_record_new_from_blob (GByteArray *blob)
{
	DmapdDPAPRecord *record = DMAPD_DPAP_RECORD (g_object_new (TYPE_DMAPD_DPAP_RECORD, NULL));

	guint8 *ptr = blob->data;

	record->priv->filesize = *((gint *) ptr);
	ptr += sizeof (record->priv->filesize);

	record->priv->largefilesize = *((gint *) ptr);
	ptr += sizeof (record->priv->largefilesize);

	record->priv->location = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	record->priv->filename = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	record->priv->thumbnail_size = *((gint *) ptr);
	ptr += sizeof (record->priv->thumbnail_size);

	record->priv->thumbnail = g_memdup (ptr, record->priv->thumbnail_size);
	ptr += record->priv->thumbnail_size;

	record->priv->aspectratio = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	record->priv->height = *((gint *) ptr);
	ptr += sizeof (record->priv->height);

	record->priv->width = *((gint *) ptr);
	ptr += sizeof (record->priv->width);

	record->priv->format = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	record->priv->comments = g_strdup ((char *) ptr);
	ptr += strlen ((char *) ptr) + 1;

	return record;
}

DmapdDPAPRecord *dmapd_dpap_record_new (const char *path)
{
	MagickWand *wand;
	DmapdDPAPRecord *record = NULL;
	wand = NewMagickWand ();

	g_warning ("Calling MagickReadImage...");
	if (! MagickReadImage (wand, path)) {
		g_warning ("could not read %s", path);
	} else {
		ExifData *d;
		struct stat buf;
		size_t size;
		float aspectratio;

		g_warning ("MagickReadImage done");
		record = DMAPD_DPAP_RECORD (g_object_new (TYPE_DMAPD_DPAP_RECORD, NULL));

		if (stat (path, &buf) == -1) {
			g_warning ("Unable to determine size of %s", path);
		} else {
			record->priv->largefilesize = buf.st_size;
		}

		record->priv->creationdate = 1;
		record->priv->rating = 5;
		record->priv->location = g_strdup_printf ("file://%s", path);
		record->priv->filename = g_basename (path);
		record->priv->title = "FIXME";

		MagickResetIterator (wand);

		record->priv->format = MagickGetImageFormat (wand);
		record->priv->height = MagickGetImageHeight (wand);
		record->priv->width = MagickGetImageWidth (wand);
		record->priv->comments = g_strdup ("");

		aspectratio = record->priv->width / (float) record->priv->height;
		record->priv->aspectratio = g_strdup_printf ("%f", aspectratio);

		g_warning ("Calculating thumbnail...");
		if (record->priv->width > record->priv->height) {
			MagickResizeImage (wand,
					   106,
					   (1 / aspectratio) * 106,
					   LanczosFilter,
					   1.0);
		} else {
			MagickResizeImage (wand,
					   aspectratio * 106,
					   106,
					   LanczosFilter,
					   1.0);
		}
		record->priv->thumbnail = MagickWriteImageBlob (wand, &size);
		record->priv->thumbnail_size = record->priv->filesize = size;
		/* FIXME:
		d = exif_data_new_from_file (path);
		if (! d) {
			g_warning ("Failed to EXIF data from %s", path);
		} else {
			int i;
			for (i = 0; i < EXIF_IFD_COUNT; i++) {
				ExifContent *c = d->ifd[i];
				if (! c || ! c->count) {
					g_warning ("Failed to find EXIF content in %s", path);
				} else {
					ExifEntry *e = exif_content_get_entry (c, EXIF_TAG_USER_COMMENT);
					if (! e) {
						g_warning ("Failed to get comments EXIF entry in %s", path);
					} else {
						gchar v[BUFSIZ + 1];
						exif_entry_get_value (e, v, BUFSIZ);
						record->priv->comments = g_strdup (v);
					exif_content_unref (c);
					}
				}
			exif_data_unref (d);
			}
		}
		*/
	}

	DestroyMagickWand (wand); 

	return record;
}
