/*   FILE: dmapd.c -- A DMAP server
 * AUTHOR: W. Michael Petullo <mike@flyn.org>
 *   DATE: 01 January 2009 
 *
 * Copyright (c) 2009 W. Michael Petullo <new@flyn.org>
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <glib/gprintf.h>
#include <glib/gdir.h>
#include <glib/gfileutils.h>
#include <libdmapsharing/dmap-mdns-publisher.h>
#include <libdmapsharing/dpap-share.h>
#include <libdmapsharing/daap-share.h>

#include "dmapd-dmap-container-record.h"
#include "dmapd-dmap-container-db.h"
#include "dmapd-dmap-db.h"
#include "dmapd-dpap-record.h"
#include "dmapd-dpap-record-factory.h"
#include "dmapd-daap-record.h"
#include "dmapd-daap-record-factory.h"

/* For use when deciding whether to serve DAAP or DPAP. */
enum {
	DAAP,
	DPAP
};

static GSList *music_dirs = NULL;
static GSList *picture_dirs = NULL;
static gboolean foreground = 0;
static gchar *share_name = NULL;

static gboolean
add_music_dir (const gchar *option_name,
			const gchar *value,
			gpointer data,
			GError **error)
{
	music_dirs = g_slist_append (music_dirs, g_strdup (value));
}

static gboolean
add_picture_dir (const gchar *option_name,
			  const gchar *value,
			  gpointer data,
			  GError **error)
{
	picture_dirs = g_slist_append (picture_dirs, g_strdup (value));
}

static GOptionEntry entries[] = {
	{ "music-dir", 'm', 0, G_OPTION_ARG_CALLBACK, add_music_dir, "Music directory", NULL },
	{ "picture-dir", 'p', 0, G_OPTION_ARG_CALLBACK, add_picture_dir, "Picture directory", NULL },
	{ "foreground", 'f', 0, G_OPTION_ARG_NONE, &foreground, "Do not fork; remain in foreground", NULL },
	{ "name", 'n', 0, G_OPTION_ARG_STRING, &share_name, "Name of media shares", NULL },
	{ NULL }
};

static gint
add_file_to_db (const char *path, DMAPDb *db, DMAPRecordFactory *factory)
{
	gint id = 0;
	DMAPRecord *record = dmap_record_factory_create (factory, g_strdup_printf ("%s", path));
	if (! record) {
		g_warning ("Could not load %s", path);
	} else {
		id = dmapd_dmap_db_add (db, record);
	}

	return id;
}

static void
build_db_starting_at (const char *dir,
			   DMAPDb *db,
			   DMAPContainerDb *container_db,
			   DMAPContainerRecord *container_record,
			   DMAPRecordFactory *factory)
{
	GError *error = NULL;

	GDir *d = g_dir_open (dir, 0, &error);

	if (error != NULL) {
		g_error ("Error opening %s", dir);
	} else {
		const gchar *entry;
		
		while (entry = g_dir_read_name (d)) {
			gchar *path = g_strdup_printf ("%s/%s", dir, entry);

			if (g_file_test (path, G_FILE_TEST_IS_DIR)) {
				g_warning ("Search %s", path);
				DMAPContainerRecord *record = dmapd_dmap_container_record_new (g_strdup (entry));
				build_db_starting_at (path, db, container_db, record, factory);
				dmapd_dmap_container_db_add (container_db, record);
			} else {
				gint id;

				id = add_file_to_db (path, db, factory);
				if (id) {
					DMAPRecord *record = dmapd_dmap_db_lookup_by_id (db, id);
					if (container_record) {
						DMAPContainerEntry *entry = g_new (DMAPContainerEntry, 1);
						entry->media_id = id;
						entry->container_id = dmapd_dmap_container_record_get_id (container_record);
						dmapd_dmap_container_record_add_entry (container_record, entry);
					}
				}
			}
			g_free (path);
		}
		
		g_dir_close (d);
	}
}

static char *
default_share_name ()
{
        const gchar *real_name;

        real_name = g_get_real_name ();
        if (strcmp (real_name, "Unknown") == 0) {
                real_name = g_get_user_name ();
        }

        return g_strdup_printf ("%s's Media", real_name);

	g_free (real_name);
}

static void
create_share (gint protocol, DMAPDb *db, DMAPContainerDb *container_db)
{
	char *name;

	DMAPShare *share = NULL;

	name = share_name ? share_name : default_share_name ();

	if (protocol == DAAP) {
		g_warning ("Initializing DAAP sharing (%d)", protocol);
		share = DMAP_SHARE (daap_share_new (name, NULL, db, container_db));
	} else {
		g_warning ("Initializing DPAP sharing (%d)", protocol);
		share = DMAP_SHARE (dpap_share_new (name, NULL, db, container_db));
	}

	g_free (name);
}

static void
daemonize (void)
{
	int child, fd;
	char *pid;

	if (getppid () == 1)
		return; /* Already a daemon */

	child = fork ();

	if (child < 0)
		g_error ("Error forking");
	
	if (child > 0)
		exit (EXIT_SUCCESS);
	
	setsid ();

	chdir (RUNDIR);

	fd = open (LOCKPATH, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

	if (lockf (fd, F_TLOCK, 0) < 0)
		g_error ("Error opening lock file");
	
	fd = open (RUNDIR "/dmapd.pid", O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

	pid = g_strdup_printf ("%d", getpid ());
	write (fd, pid, strlen (pid));
	g_free (pid);
}

static pid_t
wait_on (pid_t pid, int protocol)
{
	int status;
	waitpid (pid, &status, 0);
	if (WIFEXITED (status)) {
		return WEXITSTATUS (status) ? 1 << protocol : EXIT_SUCCESS;
	} else {
		return EXIT_FAILURE;
	}
}

int main (int argc, char *argv[])
{
	int exitval = EXIT_SUCCESS;
	pid_t child[2] = { 0, 0 };
	char *root;
	gint protocol;
	GError *error = NULL;
	GOptionContext *context;

	g_type_init ();
	g_thread_init (NULL);

	context = g_option_context_new ("- serve media using DMAP");
	g_option_context_add_main_entries (context, entries, NULL);
	g_option_context_add_group (context, gst_init_get_option_group ());
	if (! g_option_context_parse (context, &argc, &argv, &error)) {
		g_error ("Option parsing failed: %s\n", error->message);
	}

	if (! foreground) {
		daemonize ();
	}

	if (music_dirs) {
		/* FIXME: This code is redundant with DPAP code. */
		if ((child[DAAP] = fork ()) == 0) {
			GSList *l;
			DMAPDb *db;
			DMAPRecordFactory *factory;
			DMAPContainerDb *container_db;
			static GMainLoop *loop;

			factory = DMAP_RECORD_FACTORY (dmapd_daap_record_factory_new ());
			db = DMAP_DB (dmapd_dmap_db_new ());
			container_db = DMAP_CONTAINER_DB (dmapd_dmap_container_db_new ());

			for (l = music_dirs; l; l = l->next) {
				build_db_starting_at (l->data, db, container_db, NULL, factory);
			}

			loop = g_main_loop_new (NULL, FALSE);
			create_share (DAAP, DMAP_DB (db), DMAP_CONTAINER_DB (container_db));

			g_main_loop_run (loop);

			exit (EXIT_SUCCESS);
		}
	}

	if (picture_dirs) {
		if ((child[DPAP] = fork ()) == 0) {
			GSList *l;
			DMAPDb *db;
			DMAPRecordFactory *factory;
			DMAPContainerDb *container_db;
			static GMainLoop *loop;

			factory = DMAP_RECORD_FACTORY (dmapd_dpap_record_factory_new ());
			db = DMAP_DB (dmapd_dmap_db_new ());
			container_db = DMAP_CONTAINER_DB (dmapd_dmap_container_db_new ());

			for (l = picture_dirs; l; l = l->next) {
				build_db_starting_at (l->data, db, container_db, NULL, factory);
			}

			loop = g_main_loop_new (NULL, FALSE);
			create_share (DPAP, DMAP_DB (db), DMAP_CONTAINER_DB (container_db));

			g_main_loop_run (loop);

			exit (EXIT_SUCCESS);
		}
	}

	if (child[DAAP]) {
		exitval |= wait_on (child[DAAP], DAAP);
	}

	if (child[DPAP]) {
		exitval |= wait_on (child[DPAP], DPAP);
	}

	exit(exitval);
}
