/*
 * GGstInputStream class: Open a URI using g_gst_input_stream_new ().
 * Data is decoded using GStreamer and is then made available by the class's
 * read operations.
 *
 * Copyright (C) 2008 W. Michael Petullo <mike@flyn.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gst/gst.h>
#include <gst/app/gstappsink.h>
#include <gst/app/gstappbuffer.h>

#include "g-gst-input-stream.h"
#include "util.h"

struct GGstInputStreamPrivate {
	GQueue *buffer;
	gsize read_request;		/* Size of data asked for */
	GCond *buffer_ready;		/* Signals when buffer >= read_req. */
	GMutex *buffer_mutex; 		/* Protects buffer and read_request */
	gboolean buffer_closed;		/* May close before decoding complete */
	GstElement *pipeline;
	GstElement *src;
	GstElement *decode;
	GstElement *convert;
	GstElement *encode;
	GstElement *sink;
};

static void
new_decoded_pad_cb (GstElement *element,
		    GstPad *pad,
		    gboolean arg,
		    GGstInputStream *stream)
{
	/* Link remaining pad after decodebin does its magic. */
	GstPad *conv_pad;

	conv_pad = gst_element_get_static_pad (stream->priv->convert, "sink");
	g_assert (conv_pad != NULL);

	if (pads_compatible (pad, conv_pad)) {
		g_assert (! GST_PAD_IS_LINKED (
			gst_element_get_static_pad (stream->priv->convert, "sink")));

		gst_pad_link (pad, conv_pad);

		gst_element_link_many (stream->priv->convert,
				       stream->priv->encode,
				       stream->priv->sink,
				       NULL);
	} else {
		g_warning ("Could not link GStreamer pipeline.");
	}
}

static void
new_buffer_cb (GstElement *element, GGstInputStream *stream)
{
	gsize i;
	guint8 *ptr;
	GstBuffer *buffer;

	/* FIXME: Is this necessary? I am trying to protect against this
	 * thread manipulating data after the pipeline has been destroyed.
	 * see also g_gst_input_stream_close ().
	 */
	g_mutex_lock (stream->priv->buffer_mutex);

	if (stream->priv->buffer_closed) {
		g_warning ("Unread data");
		goto _return;
	}

	buffer = gst_app_sink_pull_buffer (GST_APP_SINK (element));

	if (buffer == NULL) {
		g_warning ("Error getting GStreamer buffer");
		goto _return;
	}

	ptr = GST_BUFFER_DATA (buffer);

	for (i = 0; i < GST_BUFFER_SIZE (buffer); i++) {
		g_queue_push_tail (stream->priv->buffer, GINT_TO_POINTER((gint) *ptr++));
	}

	if (g_queue_get_length (stream->priv->buffer)
	    >= stream->priv->read_request)
		g_cond_signal (stream->priv->buffer_ready);

	gst_buffer_unref (buffer);

_return:
	g_mutex_unlock (stream->priv->buffer_mutex);
}

GInputStream* g_gst_input_stream_new (gchar *uri)
{
	GstStateChangeReturn sret;
	GstState state;
	GGstInputStream *stream;
	GError *error;

	stream = G_GST_INPUT_STREAM (g_object_new (TYPE_G_GST_INPUT_STREAM,
						   NULL));

	stream->priv->buffer = g_queue_new ();
	stream->priv->read_request = 0;
	stream->priv->buffer_ready = g_cond_new ();
	stream->priv->buffer_mutex = g_mutex_new ();
	stream->priv->buffer_closed = FALSE;

	stream->priv->pipeline = gst_pipeline_new ("pipeline");

	stream->priv->src     = gst_element_factory_make ("giosrc", "src");
	stream->priv->decode  = gst_element_factory_make ("decodebin", "decode");
	stream->priv->convert = gst_element_factory_make ("audioconvert", "convert");
	stream->priv->encode  = gst_element_factory_make ("lame", "encode");
	stream->priv->sink    = gst_element_factory_make ("appsink", "sink");

	gst_bin_add_many (GST_BIN (stream->priv->pipeline),
			  stream->priv->src,
			  stream->priv->decode,
			  stream->priv->convert,
			  stream->priv->encode,
			  stream->priv->sink,
			  NULL);

	gst_element_link_many (stream->priv->src,
			       stream->priv->decode,
			       NULL);

	g_object_set (G_OBJECT (stream->priv->src), "location", uri, NULL);

	/* quality=9 is important for fast, realtime transcoding: */
	g_object_set (G_OBJECT (stream->priv->encode), "quality", 9, NULL);
	g_object_set (G_OBJECT (stream->priv->encode), "bitrate", 128, NULL);
	/* Turn off VBR: */
	g_object_set (G_OBJECT (stream->priv->encode), "vbr", 0, NULL);

	g_object_set (G_OBJECT (stream->priv->sink), "emit-signals", TRUE, "sync", FALSE, NULL);

	g_signal_connect (stream->priv->sink, "new-buffer", G_CALLBACK (new_buffer_cb), stream);
	g_signal_connect (stream->priv->decode, "new-decoded-pad", G_CALLBACK (new_decoded_pad_cb), stream);

	/* FIXME: this technique is shared with dmapd-daap-share.c */
	sret = gst_element_set_state (stream->priv->pipeline, GST_STATE_PLAYING);
	if (GST_STATE_CHANGE_ASYNC == sret) {
		if (GST_STATE_CHANGE_SUCCESS != gst_element_get_state (GST_ELEMENT (stream->priv->pipeline), &state, NULL, 5 * GST_SECOND)) {
			g_warning ("State change failed for %s.", uri);
		}
	} else if (sret != GST_STATE_CHANGE_SUCCESS) {
		g_print ("%s - Could not read file %d\n", uri, sret);
	}

	return G_INPUT_STREAM (stream);
}

GInputStream* g_gst_input_stream_new_sized (gchar *uri, gsize size)
{
	return g_gst_input_stream_new (uri);
}

static gssize
g_gst_input_stream_read (GInputStream  *stream,
			 void          *buffer,
			 gsize          count,
			 GCancellable  *cancellable,
			 GError       **error)
{
	int i;
	GGstInputStream *gst_stream = G_GST_INPUT_STREAM (stream);
	GTimeVal time;

	g_get_current_time (&time);
	g_time_val_add (&time, 1000000); /* 1 second in microseconds */

	g_mutex_lock (gst_stream->priv->buffer_mutex);

	gst_stream->priv->read_request = count;
	if (g_queue_get_length (gst_stream->priv->buffer) < count
	    && ! g_cond_timed_wait (gst_stream->priv->buffer_ready,
			   gst_stream->priv->buffer_mutex,
			   &time)) {
		/* Timeout: Count is now what's remaining.  Let's hope
		 * we have enough of a lead on encoding so that this one
		 * second timeout will go unnoticed.
		 */
		g_warning ("Timeout waiting for converted data");
		count = g_queue_get_length (gst_stream->priv->buffer);
	}

	for (i = 0; i < count; i++) {
		((guint8 *) buffer)[i] = GPOINTER_TO_INT (
			g_queue_pop_head (gst_stream->priv->buffer));
	}
		
	g_mutex_unlock (gst_stream->priv->buffer_mutex);

	return count;
}

static gssize
g_gst_input_stream_skip (GInputStream *stream,
			    gsize count,
			    GCancellable *cancellable,
			    GError **error)
{
}

static gboolean
g_gst_input_stream_close (GInputStream *stream,
			  GCancellable *cancellable,
			  GError **error)
{
	GGstInputStream *gst_stream = G_GST_INPUT_STREAM (stream);

	g_mutex_lock (gst_stream->priv->buffer_mutex);

	gst_element_set_state (gst_stream->priv->pipeline, GST_STATE_NULL);
	gst_object_unref (GST_OBJECT (gst_stream->priv->pipeline));
	g_queue_free (gst_stream->priv->buffer);

	gst_stream->priv->buffer_closed = TRUE;

	g_mutex_unlock (gst_stream->priv->buffer_mutex);
}

static gssize
g_gst_input_stream_read_finish (GInputStream *stream,
				GAsyncResult *result,
				GError **error)
{
}

static gssize
g_gst_input_stream_skip_finish (GInputStream *stream,
				GAsyncResult *result,
				GError **error)
{
}

static void
g_gst_input_stream_close_async (GInputStream *stream,
				int io_priority,
				GCancellable *cancellabl,
				GAsyncReadyCallback callback,
				gpointer data)
{
}

static void
g_gst_input_stream_read_async (GInputStream *stream,
			       void *buffer,
			       gsize count,
			       int io_priority,
			       GCancellable *cancellable,
			       GAsyncReadyCallback callback,
			       gpointer user_data)
{
}

static void
g_gst_input_stream_skip_async (GInputStream *stream,
			       gsize count,
			       int io_priority,
			       GCancellable *cancellabl,
			       GAsyncReadyCallback callback,
			       gpointer datae)
{
}

static gboolean
g_gst_input_stream_close_finish (GInputStream *stream,
			         GAsyncResult *result,
				 GError **error)
{
}

static void
g_gst_input_stream_class_init (GGstInputStreamClass *klass)
{
	GObjectClass *object_class;
	GInputStreamClass *istream_class;

	g_type_class_add_private (klass, sizeof (GGstInputStreamPrivate));

	istream_class = G_INPUT_STREAM_CLASS (klass);
	istream_class->read_fn      = g_gst_input_stream_read;
	istream_class->skip         = g_gst_input_stream_skip;
	istream_class->close_fn     = g_gst_input_stream_close;
	istream_class->read_async   = g_gst_input_stream_read_async;
	istream_class->read_finish  = g_gst_input_stream_read_finish;
	istream_class->skip_async   = g_gst_input_stream_skip_async;
	istream_class->skip_finish  = g_gst_input_stream_skip_finish;
	istream_class->close_async  = g_gst_input_stream_close_async;
	istream_class->close_finish = g_gst_input_stream_close_finish;
}

static void
g_gst_input_stream_init (GGstInputStream *stream)
{
	stream->priv = G_GST_INPUT_STREAM_GET_PRIVATE (stream);

}

G_DEFINE_TYPE (GGstInputStream, g_gst_input_stream, G_TYPE_INPUT_STREAM)
